import os
import json
import shutil
import argparse

from barbell2.utils import Logger
from barbell2.lib.dicom import is_dicom
from types import SimpleNamespace


class GenID:

    def __init__(self, args):
        if isinstance(args, dict):
            self.args = SimpleNamespace(**args)
        else:
            self.args = args
        self.include_types = [x.strip() for x in self.args.include_types.split(',')]
        self.logger = Logger(prefix='genID', to_dir=self.args.log_dir, timestamp=False)
        self.prefix = self.args.prefix
        self.nr_digits = self.args.nr_digits
        self.idx = 1

    @staticmethod
    def get_tag_file(dcm_file):
        tag_file = dcm_file[:-4] + '.tag'
        if os.path.isfile(tag_file):
            return tag_file
        tag_file = dcm_file + '.tag'
        if os.path.isfile(tag_file):
            return tag_file
        return None

    def next_file_id(self):
        file_id = '{}{}'.format(self.prefix, str(self.idx).zfill(self.nr_digits))
        self.idx += 1
        return file_id

    def execute(self):
        file_ids = {}
        os.makedirs(self.args.output_dir, exist_ok=False)
        for root, dirs, files in os.walk(self.args.root_dir):
            for f in files:
                f = os.path.join(root, f)
                file_id = self.next_file_id()
                if 'dicom' in self.include_types and 'tag' in self.include_types and is_dicom(f):
                    dcm_file = f
                    tag_file = self.get_tag_file(dcm_file)
                    if tag_file is None:
                        continue
                    target_dcm_file = os.path.join(self.args.output_dir, file_id + '.dcm')
                    shutil.copyfile(dcm_file, target_dcm_file)
                    target_tag_file = os.path.join(self.args.output_dir, file_id + '.tag')
                    shutil.copyfile(tag_file, target_tag_file)
                    file_ids[file_id] = [target_dcm_file, target_tag_file]
                elif 'dicom' in self.include_types:
                    dcm_file = f
                    target_dcm_file = os.path.join(self.args.output_dir, file_id + '.dcm')
                    shutil.copyfile(dcm_file, target_dcm_file)
                    file_ids[file_id] = target_dcm_file
                else:
                    pass
        with open(os.path.join(self.args.output_dir, '..', 'file_ids.txt'), 'w') as f:
            json.dump(file_ids, f, indent=4)
        return self.args.output_dir


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'root_dir',
        help='Root directory containing DICOM and (optionally) TAG files',
    )
    parser.add_argument(
        'output_dir',
        help='Directory where to write renamed files',
    )
    parser.add_argument(
        'include_types',
        help='File types to include (default: dicom,tag)',
        default='dicom,tag',
    )
    parser.add_argument(
        'prefix',
        help='Optional prefix to put in front of file index',
        default='P',
    )
    parser.add_argument(
        'nr_digits',
        help='Number of digits to use in file name',
        default=4,
    )
    parser.add_argument(
        '--log_dir',
        help='Directory where to write logging info (default: .)',
        default='.',
    )
    args = parser.parse_args()
    x = GenID(args)
    x.execute()


if __name__ == '__main__':
    main()



from autosegl3 import GenID


def test_genid():
    args = {
        'root_dir': '/Volumes/USB_SECURE1/data/radiomics/projects/deepseg/data/mega/processed',
        'output_dir': '/Volumes/USB_SECURE1/data/radiomics/projects/deepseg/data/mega/renamed',
        'include_types': 'dicom, tag',
        'prefix': 'P',
        'nr_digits': 4,
        'log_dir': '.',
    }
    x = GenID(args)
    x.execute()



import json

from types import SimpleNamespace
from autosegl3 import TrainL3


def test_train():
    with open('data/params.json', 'r') as f:
        args = SimpleNamespace(**json.load(f))
        x = TrainL3(args)
        model_path = x.execute()
        assert 'final_model' in model_path



import tensorflow as tf
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Conv2D, Input, MaxPooling2D, Dropout, \
    concatenate, BatchNormalization, Conv2DTranspose, PReLU


def conv_block(inputs, num_features, kernel_size, params):
    x = Conv2D(num_features, kernel_size, activation=None, kernel_initializer='he_normal',
               padding='same', kernel_regularizer=tf.keras.regularizers.l2(l=params.l2_loss))(inputs)
    x = PReLU(shared_axes=[1, 2])(x)
    x = BatchNormalization()(x)
    x = Dropout(params.dropout_rate)(x)
    x = Conv2D(num_features, kernel_size, activation=None, kernel_initializer='he_normal',
               padding='same', kernel_regularizer=tf.keras.regularizers.l2(l=params.l2_loss))(x)
    x = PReLU(shared_axes=[1, 2])(x)
    x = BatchNormalization()(x)
    return x


def unet(params, optimizer, loss):
    num_classes = params.num_classes
    input_ct = Input((None, None, params.patch_shape[-1]), name='CT_input')
    x_1 = conv_block(input_ct, 32, (3, 3), params)
    p_1 = MaxPooling2D((2, 2))(x_1)

    x_2 = conv_block(p_1, 64, (3, 3), params)
    p_2 = MaxPooling2D((2, 2))(x_2)

    x_3 = conv_block(p_2, 128, (3, 3), params)
    p_3 = MaxPooling2D((2, 2))(x_3)

    x_4 = conv_block(p_3, 256, (3, 3), params)
    p_4 = MaxPooling2D((2, 2))(x_4)

    x_5 = conv_block(p_4, 512, (3, 3), params)

    u6 = Conv2DTranspose(128, (2, 2), strides=(2, 2), padding='same')(x_5)
    u6 = concatenate([u6, x_4])
    x_6 = conv_block(u6, 256, (3, 3), params)

    u7 = Conv2DTranspose(64, (2, 2), strides=(2, 2), padding='same')(x_6)
    u7 = concatenate([u7, x_3])
    x_7 = conv_block(u7, 128, (3, 3), params)

    u8 = Conv2DTranspose(32, (2, 2), strides=(2, 2), padding='same')(x_7)
    u8 = concatenate([u8, x_2])
    x_8 = conv_block(u8, 64, (3, 3), params)

    u9 = Conv2DTranspose(16, (2, 2), strides=(2, 2), padding='same')(x_8)
    u9 = concatenate([u9, x_1])
    x_9 = conv_block(u9, 32, (3, 3), params)

    outputs = Conv2D(num_classes, (1, 1), activation='softmax')(x_9)

    model = Model(inputs=[[input_ct]], outputs=outputs)
    model.compile(optimizer=optimizer, loss=loss, metrics=['accuracy'])

    return model



import os
import datetime
import numpy as np
import tensorflow as tf

from tensorflow.keras import losses
from types import SimpleNamespace
from barbell2.utils import Logger
from autosegl3 import models
from autosegl3 import utils


def run_once(f):
    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.has_run = True
            return f(*args, **kwargs)

    wrapper.has_run = False
    return wrapper


@run_once
def start_graph_tensorflow():
    tf.summary.trace_on(graph=True, profiler=True)


@run_once
def end_graph_tensorflow(self, log_dir):
    with self.as_default():
        tf.summary.trace_export(name="graph", step=0, profiler_outdir=log_dir)


def dice_loss(y_true, y_pred):
    smooth = 1
    union = tf.reduce_sum(y_true, axis=[1, 2, 3]) + tf.reduce_sum(y_pred, axis=[1, 2, 3])
    intersection = tf.reduce_sum(y_true * y_pred, axis=[1, 2, 3])
    score = tf.reduce_mean((2. * intersection + smooth) / (union + smooth), axis=0)
    return 1 - score


def early_stopping(loss_list, min_delta=0.005, patience=20):
    if len(list(loss_list)) // patience < 2:
        return False
    mean_previous = np.mean(loss_list[::-1][patience:2 * patience])
    mean_recent = np.mean(loss_list[::-1][:patience])
    delta_abs = np.abs(mean_recent - mean_previous)  # abs change
    delta_abs = np.abs(delta_abs / mean_previous)  # relative change
    if delta_abs < min_delta:
        print('Stopping early...')
        return True
    else:
        return False


def get_batch(images, labels, params):
    ct_batch = np.zeros(shape=[params['batch_size'],
                               params['patch_shape'][0],
                               params['patch_shape'][1],
                               params['patch_shape'][2]])
    gt_batch = np.zeros(shape=[params['batch_size'],
                               params['patch_shape'][0],
                               params['patch_shape'][1],
                               params['patch_shape'][2]])
    for patch in range(0, params['batch_size']):
        patient = random.randint(0, images.shape[-1] - 1)
        ct = images[:, :, patient]
        gt = labels[:, :, patient]
        # ct = utils.normalize(ct, 'True', params['min_bound'], params['max_bound'])
        # if random.randint(0, 1) == 1:
        #    num_augments = np.random.randint(1, params['number_of_augmentations'] + 1)
        #    ct, gt = data_augmentation.apply_augmentations(ct,
        #                                                   gt,
        #                                                   num_augments)
        ct_batch[patch, :, :, 0] = ct
        gt_batch[patch, :, :, 0] = gt
    gt_batch = tf.one_hot(np.uint8(np.squeeze(gt_batch, axis=-1)), params['num_classes'])
    return ct_batch, gt_batch


class DiceMetric(tf.keras.metrics.Metric):

    def __init__(self, name='dice_coefficient', **kwargs):
        super(DiceMetric, self).__init__(name=name, **kwargs)
        self.dice_score = self.add_weight(name='dsc', initializer='zeros')

    def update_state(self, y_true, y_pred, sample_weight=None):
        smooth = 0.000001
        union = tf.reduce_sum(y_true, axis=[1, 2, 3]) + tf.reduce_sum(y_pred, axis=[1, 2, 3])
        intersection = tf.reduce_sum(y_true * y_pred, axis=[1, 2, 3])
        score = tf.reduce_mean((2. * intersection) / (union + smooth), axis=0)
        self.dice_score.assign(score)

    def result(self):
        return self.dice_score

    def reset_states(self):
        self.dice_score.assign(0.0)


class TrainL3:

    def __init__(self, args):
        if isinstance(args, dict):
            self.params = SimpleNamespace(**args)
        else:
            self.params = args
        self.logger = Logger(prefix='trainl3', to_dir='.')
        self.train_test_split = True if self.params.test_size > 0.0 else False

    def execute(self):

        @tf.function
        def train_on_batch(im_src, gt_src):
            with tf.GradientTape() as tape:
                predictions = model(inputs=[im_src], training=True)
                regularization_loss = tf.math.add_n(model.losses)
                loss_value = loss_function(gt_src, predictions)
                total_loss = regularization_loss + loss_value
            grads = tape.gradient(total_loss, model.trainable_weights)
            optimizer_function.apply_gradients(zip(grads, model.trainable_weights))
            train_loss(total_loss)
            return predictions

        @tf.function
        def validate_on_batch(im_src, gt_src):
            predictions = model(inputs=[im_src], training=False)
            regularization_loss = tf.math.add_n(model.losses)
            loss_value = loss_function(gt_src, predictions)
            total_loss = regularization_loss + loss_value
            validation_loss(total_loss)
            return predictions

        loss_list = []
        loss_function = losses.CategoricalCrossentropy()
        optimizer_function = tf.keras.optimizers.Adam(self.params.learning_rate)

        final_model_path = None

        for iteration in range(0, self.params.num_steps + 1):

            # Define model
            model = models.unet(self.params,
                                optimizer=optimizer_function,
                                loss=loss_function)

            # Define evaluation metrics
            train_loss = tf.keras.metrics.Mean(name='train_loss')
            train_accuracy = DiceMetric()
            validation_loss = tf.keras.metrics.Mean(name='validation_loss')
            validation_accuracy = DiceMetric()

            # Create variables for various paths used for storing training information
            current_time = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
            if not os.path.exists(self.params.output_dir):
                os.mkdir(self.params.output_dir)

            train_log_dir = self.params.output_dir + '/gradient_tape/' + current_time + '/train'
            val_log_dir = self.params.output_dir + '/gradient_tape/' + current_time + '/val'
            saved_model_path = self.params.output_dir + '/gradient_tape/' + current_time + '/saved_models/'
            saved_weights_path = self.params.output_dir + '/gradient_tape/' + current_time + '/saved_weights/'

            train_summary_writer = tf.summary.create_file_writer(train_log_dir)
            val_summary_writer = tf.summary.create_file_writer(val_log_dir)

            os.mkdir(saved_model_path)
            os.mkdir(saved_weights_path)

            images_t, labels_t, names_t = utils.load_dataset(self.params.data_path_train, self.params)
            print('Training set size: ', np.shape(images_t)[2])

            # Start training loop
            start_graph_tensorflow()

            ct_batch, gt_batch = self.get_batch(images_t, labels_t, self.params)
            train_pred = train_on_batch(ct_batch, gt_batch)

            end_graph_tensorflow(train_summary_writer, train_log_dir)

            # Evaluation step during training.
            if iteration % self.params.train_eval_step == 0:
                # Write training information to training log
                with train_summary_writer.as_default():
                    train_dice = train_accuracy(gt_batch, train_pred).numpy()
                    tf.summary.scalar('loss', train_loss.result(), step=iteration)
                    tf.summary.scalar('accuracy', train_dice, step=iteration)

                template = 'Iteration {}, Loss: {:.5}, Dice: {:.5}'
                print(template.format(iteration + 1,
                                      train_loss.result(),
                                      train_dice))

            # Evaluation step for validation.
            # TODO: Figure out best way to call train.py with train/test split or not
            if self.train_test_split:
                images_v, labels_v, names_v = utils.load_dataset(self.params.data_path_val, self.params)
                print('Test set size: ', np.shape(images_v)[2])
                if iteration % self.params.val_eval_step == 0:
                    ct_batch_val, gt_batch_val = self.get_batch(images_v, labels_v, self.params)

                    # gt_batch_val = tf.one_hot(np.uint8(np.squeeze(gt_batch_val)), params.dict['num_classes'])
                    val_pred = validate_on_batch(ct_batch_val, gt_batch_val)

                    # Write validation information to log
                    with val_summary_writer.as_default():
                        validation_dice = validation_accuracy(gt_batch_val, val_pred).numpy()
                        tf.summary.scalar('loss', validation_loss.result(), step=iteration)
                        tf.summary.scalar('accuracy', validation_dice, step=iteration)
                        loss_list.append(validation_loss.result())

                    template = 'Iteration {}, Validation Loss: {:.5}, Validation Dice: {:.5}'
                    print(template.format(iteration + 1,
                                          validation_loss.result(),
                                          validation_dice))

                    # Earling stopping when loss in the past 'patience' train_eval_steps
                    # is smaller than 'min_delta'. Breaks loop.
                    early_stop = self.early_stopping(loss_list, min_delta=0.001, patience=10)
                    if early_stop:
                        print("Early stopping signal received at iteration = %d/%d" % (iteration, self.params.num_steps))
                        print("Terminating training ")
                        model.save(os.path.join(saved_model_path, 'model_' + str(iteration)))
                        model.save_weights(os.path.join(saved_weights_path, 'model_weights' + str(iteration) + '.h5'))
                        final_model = model
                        final_model_path = os.path.join(saved_model_path, 'final_model')
                        final_model.save(final_model_path)
                        final_model.save_weights(os.path.join(saved_weights_path, 'final_model_weights' + '.h5'))
                        break

            # Save the model at predefined step numbers.
            if iteration % self.params.save_model_step == 0:
                model.save(os.path.join(saved_model_path, 'model_' + str(iteration)))
                model.save_weights(os.path.join(saved_weights_path, 'model_weights' + str(iteration) + '.h5'))

        return final_model_path



import os


def get_available():

    # Generate output nvidia-smi
    file_path = '{}/nvidia-smi.txt'.format(os.environ['HOME'])
    cmd = 'nvidia-smi > {}'.format(file_path)
    os.system(cmd)

    # In the following I'm assuming that the GeForce cards are listed in order
    # meaning that the first GeForce encountered gets ID = 0, the second ID = 1
    # So even though nvidia-smi reports the 2nd GeForce card to have ID = 2, this
    # code will list it as ID = 1. This seems to work for now.
    memory_usage = {}
    with open(file_path, 'r') as f:
        idx = 0
        for line in f.readlines():
            if 'GeForce RTX' in line:
                memory_usage[idx] = 0
                continue
            if idx in memory_usage.keys() and memory_usage[idx] == 0:
                memory_usage[idx] = int(line.split()[12][:-1])
                idx += 1

    # Find GPU device ID with least amount of memory usage
    available_gpu = -1
    min_mu = 100
    for gpu in memory_usage.keys():
        if memory_usage[gpu] < min_mu:
            min_mu = memory_usage[gpu]
            available_gpu = gpu
    return available_gpu


if __name__ == '__main__':
    print(get_available())



import os

from autosegl3 import AutoSegL3


def test_autosegl3():
    params = {
        'image_shape': [512, 512, 1],
        'patch_shape': [512, 512, 1],
        'output_dir': '/tmp/auto_seg',
        'test_size': 0.2,
        'learning_rate': 0.001,
        'l2_loss': 0.001,
        'num_steps': 100000,
        'num_classes': 4,
        'dropout_rate': 0.05,
        'min_bound': -200,
        'max_bound': 200,
        'batch_size': 1,
        'train_eval_step': 100,
        'val_eval_step': 100,
        'save_model_step': 1000,
    }
    d = '{}/data/surfdrive/projects/20210203_autosegl3/dicom_and_tag'.format(os.environ['HOME'])
    tool = AutoSegL3(params=params)
    print(tool.train(d))
